<?php
declare(strict_types=1);
require_once __DIR__.'/inc/header.php';

class ArmorStore {
    private PDO $db;
    private User $user;
    private array $errors = [];
    private string $csrfToken;
    
    public function __construct(PDO $db, User $user) {
        $this->db = $db;
        $this->user = $user;
        $this->csrfToken = csrf_create('csrfg', false);
    }

    private function getRotatingItemSeed(): int {
        // Create a predictable but rotating selection based on current hour
        return intval(date('G')); // 0-23 hour
    }

    private function getAvailableArmor(): array {
        // Get all possible armor items
        $stmt = $this->db->prepare('
            SELECT id, description, image, name, cost, defense 
            FROM items 
            WHERE defense > 0 AND buyable = 1 
            ORDER BY defense
        ');
        $stmt->execute();
        $allArmor = $stmt->fetchAll(PDO::FETCH_ASSOC);
        
        // Rotate available items based on hour
        $seed = $this->getRotatingItemSeed();
        $itemCount = count($allArmor);
        $displayCount = min(8, $itemCount); // Show max 8 items at a time
        
        // Use seed to determine which items to show
        $selectedIndexes = array_slice(
            array_unique(
                array_map(
                    fn($i) => ($seed + $i * 7) % $itemCount, 
                    range(0, $displayCount - 1)
                )
            ),
            0,
            $displayCount
        );
        
        return array_map(fn($i) => $allArmor[$i], $selectedIndexes);
    }

    private function handlePurchase(): ?string {
        $buyId = $_GET['buy'] ?? null;
        if (!$buyId || !ctype_digit($buyId)) {
            return null;
        }

        if (!csrf_check('csrfg', $_GET)) {
            return Message(SECURITY_TIMEOUT_MESSAGE);
        }

        $stmt = $this->db->prepare('
            SELECT id, name, cost 
            FROM items 
            WHERE id = ? AND buyable = 1
        ');
        $stmt->execute([$buyId]);
        
        if (!$stmt->rowCount()) {
            $this->errors[] = 'The item you selected doesn\'t exist';
            return null;
        }

        $item = $stmt->fetch(PDO::FETCH_ASSOC);
        $itemPopup = item_popup($item['id'], $item['name']);

        if ($item['cost'] > $this->user->money) {
            $this->errors[] = 'You don\'t have enough money to buy '.aAn($item['name'], false).$itemPopup;
            return null;
        }

        try {
            $this->db->beginTransaction();
            
            $stmt = $this->db->prepare('
                UPDATE users 
                SET money = GREATEST(money - ?, 0) 
                WHERE id = ?
            ');
            $stmt->execute([$item['cost'], $this->user->id]);
            
            Give_Item($buyId, $this->user->id);
            
            $this->db->commit();
            return Message('You\'ve purchased '.aAn($item['name'], false).$itemPopup);
            
        } catch (Exception $e) {
            $this->db->rollBack();
            $this->errors[] = 'Transaction failed: ' . $e->getMessage();
            return null;
        }
    }

    public function render(): void {
        $purchaseMessage = $this->handlePurchase();
        $availableArmor = $this->getAvailableArmor();
        
        if (empty($availableArmor)) {
            ?>
            </table><!-- Close table from header -->
            <main class="store-container">
                <?php echo Message('There are currently no armors available', 'Error', true); ?>
            </main>
            <?php
            return;
        }
        // Start HTML output
        ?>
        </table><!-- Close table from header -->

<main class="store-container">
    <div class="store-header-wrapper">
        <header class="store-header">
            <h1 class="store-title">Crazy Riley's Armor Emporium</h1>
        </header>
        <div class="store-time">
            <svg class="store-time-icon" viewBox="0 0 24 24" width="16" height="16">
                <path fill="currentColor" d="M12 2C6.5 2 2 6.5 2 12s4.5 10 10 10 10-4.5 10-10S17.5 2 12 2zm0 18c-4.4 0-8-3.6-8-8s3.6-8 8-8 8 3.6 8 8-3.6 8-8 8zm.5-13H11v6l5.2 3.2.8-1.3-4.5-2.7V7z"/>
            </svg>
            Rotation in: <?php echo 60 - intval(date('i')); ?>m
        </div>
    </div>

    <div class="store-welcome">
        <p class="store-message">
            Welcome to Crazy Riley's Armor Emporium! Our selection rotates hourly to bring you the finest protection in all the realm.
        </p>
    </div>

    <?php if ($purchaseMessage || !empty($this->errors)): ?>
    <div class="store-notifications">
        <?php if (!empty($this->errors)): ?>
            <div class="store-errors">
                <?php display_errors($this->errors); ?>
            </div>
        <?php endif; ?>
        
        <?php if ($purchaseMessage): ?>
            <div class="store-notification">
                <?php echo $purchaseMessage; ?>
            </div>
        <?php endif; ?>
    </div>
    <?php endif; ?>

    <div class="store-grid">
        <?php 
        // Ensure exactly 12 items are shown
        $itemsToShow = array_slice($availableArmor, 0, 12);
        foreach ($itemsToShow as $armor): 
        ?>
            <article class="store-item">
                <div class="item-image">
                    <img src="<?php echo format($armor['image']); ?>" 
                         alt="<?php echo htmlspecialchars($armor['name']); ?>"
                         loading="lazy" />
                </div>
                <div class="item-details">
                    <h2 class="item-name">
                        <?php echo item_popup($armor['id'], $armor['name']); ?>
                        <p class="item-description" title="<?php echo htmlspecialchars($armor['description']); ?>">
                        <?php echo htmlspecialchars($armor['description'] ?? 'No description available.'); ?>
                    </p>
                    </h2>
                    <div class="item-stats">
                        <div class="item-price">
                            <?php echo prettynum($armor['cost'], true); ?>
                        </div>
                        <div class="item-defense">
                            DEF: <?php echo $armor['defense']; ?>
                        </div>
                    </div>
                    <a href="astore.php?buy=<?php echo $armor['id']; ?>&amp;csrfg=<?php echo $this->csrfToken; ?>" 
                       class="buy-button"
                       role="button">
                        <svg viewBox="0 0 24 24" width="16" height="16">
                            <path fill="currentColor" d="M17 18c-1.1 0-2 .9-2 2s.9 2 2 2 2-.9 2-2-.9-2-2-2zM7 18c-1.1 0-2 .9-2 2s.9 2 2 2 2-.9 2-2-.9-2-2-2zm0-3l1.1-2h7.45c.75 0 1.41-.41 1.75-1.03l3.86-7.01L19.42 4l-3.87 7H8.53L4.27 2H1v2h2l3.6 7.59L5.2 14v1c0 1.1.9 2 2 2h12v-2H7z"/>
                        </svg>
                        Buy Now
                    </a>
                </div>
            </article>
        <?php endforeach; ?>
    </div>
</main>
        <?php
    }
}

// Initialize and render the store
$store = new ArmorStore($db, $user_class);
$store->render();
?>
<style>
/* Store Container */
.store-container {
    width: 100%;
    background: var(--bg-secondary);
    border: 1px solid var(--border-color);
    border-radius: 8px;
    overflow: hidden;
}

/* Header Area */
.store-header-wrapper {
    display: flex;
    justify-content: space-between;
    align-items: stretch;
    border-bottom: 2px solid var(--accent-primary);
    background: var(--bg-accent);
}

.store-header {
    padding: 1rem 1.5rem;
    flex: 1;
}

.store-title {
    color: var(--text-primary);
    font-size: 1.25rem;
    margin: 0;
    font-weight: 600;
}

.store-time {
    padding: 1rem 1.5rem;
    color: var(--accent-primary);
    background: rgba(204, 17, 17, 0.1);
    border-left: 1px solid var(--border-color);
    font-weight: 500;
    display: flex;
    align-items: center;
    gap: 0.5rem;
    white-space: nowrap;
}

.store-time-icon {
    opacity: 0.8;
}

/* Welcome Section */
.store-welcome {
    padding: 1rem 1.5rem;
    background: linear-gradient(
        to bottom,
        var(--bg-accent),
        var(--bg-secondary)
    );
}

.store-message {
    margin: 0;
    color: var(--text-primary);
}

/* Notifications */
.store-notifications {
    padding: 1rem 1.5rem;
    background: var(--bg-secondary);
    border-bottom: 1px solid var(--border-color);
}

.store-notification,
.store-errors {
    padding: 0.75rem 1rem;
    border-radius: 4px;
    background: rgba(0, 0, 0, 0.2);
    margin-bottom: 0.5rem;
}

.store-notification:last-child,
.store-errors:last-child {
    margin-bottom: 0;
}

/* Product Grid */
.store-grid {
    display: grid;
    grid-template-columns: repeat(4, 1fr);
    gap: 1rem;
    padding: 1.5rem;
    background: var(--bg-secondary);
    max-width: 1200px;
    margin: 0 auto;
}

/* Product Cards */
.store-item {
    background: var(--bg-accent);
    border: 1px solid var(--border-color);
    border-radius: 6px;
    overflow: hidden;
    transition: all 0.2s ease;
    display: flex;
    flex-direction: column;
    width: 100%;
    min-height: 280px;
}

.store-item:hover {
    transform: translateY(-2px);
    box-shadow: 0 4px 12px rgba(0, 0, 0, 0.2);
    border-color: var(--accent-primary);
}

/* Image Container */
.item-image {
    position: relative;
    width: 100%;
    height: 140px;
    background: rgba(0, 0, 0, 0.2);
    border-bottom: 1px solid var(--border-color);
}

.item-image img {
    position: absolute;
    top: 0;
    left: 0;
    width: 100%;
    height: 100%;
    object-fit: contain;
    padding: 0.75rem;
    box-sizing: border-box;
    transition: transform 0.2s ease;
}

.store-item:hover .item-image img {
    transform: scale(1.05);
}

/* Item Details */
.item-details {
    padding: 0.75rem;
    display: flex;
    flex-direction: column;
    gap: 0.5rem;
    flex: 1;
    background: linear-gradient(
        to bottom,
        var(--bg-accent),
        var(--bg-secondary)
    );
}

.item-info {
    flex: 1;
    display: flex;
    flex-direction: column;
    gap: 0.35rem;
}

.item-name {
    font-size: 0.9rem;
    font-weight: 600;
    color: var(--text-primary);
    margin: 0;
    line-height: 1.3;
    display: -webkit-box;
    -webkit-line-clamp: 2;
    -webkit-box-orient: vertical;
    overflow: hidden;
}

.item-description {
    font-size: 0.8rem;
    color: var(--text-secondary);
    margin: 0;
    line-height: 1.4;
    display: -webkit-box;
    -webkit-line-clamp: 3;
    -webkit-box-orient: vertical;
    overflow: hidden;
    opacity: 0.8;
}

/* Stats Row */
.item-stats {
    display: flex;
    justify-content: space-between;
    align-items: center;
    gap: 0.5rem;
    margin-top: auto;
    padding-top: 0.5rem;
    border-top: 1px solid rgba(255, 255, 255, 0.1);
}

.item-price {
    color: var(--accent-primary);
    font-size: 0.95rem;
    font-weight: 600;
}

.item-defense {
    color: var(--text-secondary);
    font-size: 0.8rem;
    background: rgba(0, 0, 0, 0.2);
    padding: 0.25rem 0.5rem;
    border-radius: 3px;
    white-space: nowrap;
}

/* Buy Button */
.buy-button {
    background: var(--accent-primary);
    color: white;
    padding: 0.5rem;
    border-radius: 4px;
    text-decoration: none;
    text-align: center;
    font-weight: 500;
    font-size: 0.9rem;
    transition: all 0.2s ease;
    display: flex;
    align-items: center;
    justify-content: center;
    gap: 0.35rem;
}

.buy-button:hover {
    background: var(--accent-hover);
}

.buy-button:active {
    transform: translateY(1px);
}

/* Hover Effects */
.store-item:hover .item-description {
    color: var(--text-primary);
    opacity: 1;
}

/* Responsive Design */
@media (min-width: 1400px) {
    .item-description {
        -webkit-line-clamp: 4;
    }
}

@media (max-width: 1200px) {
    .store-grid {
        grid-template-columns: repeat(3, 1fr);
        padding: 1.25rem;
    }
    
    .item-description {
        -webkit-line-clamp: 3;
    }
}

@media (max-width: 900px) {
    .store-header-wrapper {
        flex-direction: column;
    }
    
    .store-time {
        border-left: none;
        border-top: 1px solid var(--border-color);
        justify-content: center;
    }
    
    .store-grid {
        grid-template-columns: repeat(2, 1fr);
        gap: 0.75rem;
        padding: 1rem;
    }

    .item-image {
        height: 120px;
    }
    
    .item-description {
        -webkit-line-clamp: 2;
    }
}

@media (max-width: 480px) {
    .store-grid {
        grid-template-columns: repeat(2, 1fr);
        gap: 0.5rem;
        padding: 0.75rem;
    }

    .item-image {
        height: 100px;
    }

    .item-details {
        padding: 0.5rem;
    }

    .item-name {
        font-size: 0.8rem;
        -webkit-line-clamp: 1;
    }
    
    .item-description {
        display: none;
    }

    .item-stats {
        flex-direction: column;
        align-items: flex-start;
        gap: 0.25rem;
    }

    .buy-button {
        padding: 0.4rem;
        font-size: 0.8rem;
    }
}
</style>
